//  .------------------------------------------------------------------------.
//  | FS4-MIDI-Adapter                                                       |
//  | Copyright (C) 2022-07-02 Paul Wilhelm, M. Sc. <anfrage@paulwilhelm.de> |
//  '------------------------------------------------------------------------'
//
//  Input mapping:
//    If switch <x> is OPEN,   the REMOTE<x>_IN is HIGH, supplying the footswitch LED (glowing, ON).       Then the R<x>_IN (evaluated by THIS CODE) is LOW.
//    If switch <x> is CLOSED, the REMOTE<x>_IN is LOW,  short-circuiting the LED footswitch (dark, OFF).  Then the R<x>_IN (evaluated by THIS CODE) is HIGH.
//  Simply put:
//    If the LED is ON,  the R<x>_IN is LOW!
//    If the LED is OFF, the R<x>_IN is HIGH!)
//  So:
//    The footswitch LED state is inverted with respect to the R<x>_IN input state.
//
//  Output mapping:
//    If switch <x> is OPEN,   the output transistor should also be OPEN-CIRCUIT.  This means that R<x>_OUT should be LOW.
//    If switch <x> is CLOSED, the output transistor should also be SHORT-CIRCUIT. This means that R<x>_OUT should be HIGH.
//  Simply put:
//    If R<x>_IN is LOW,  the R<x>_OUT should be LOW.
//    If R<x>_IN is HIGH, the R<x>_OUT should be HIGH.
//
//  R<x>_IN and R<x>_OUT pin map:
//                                     R1    R2    R3    R4
unsigned int  REMOTE_IN      [4] = {   12,    8,    4,    5 };
unsigned int  REMOTE_OUT     [4] = {   18,   16,   13,   14 };
//
//  Edge Detector needs (a) Timestamp of last signal change, and (b) Current state
//                                     R1    R2    R3    R4
unsigned long REMOTE_LAST_MS [4] = {    0,    0,    0,    0 };
bool          REMOTE_STATE   [4] = { HIGH, HIGH, HIGH, HIGH };  // Initial state: All R<x>_IN = HIGH means all LEDs = OFF
//
// MIDI Preset Numbers:
uint8_t PRESET_RHYTHM = 0;
uint8_t PRESET_LEAD   = 1;
uint8_t PRESET_CLEAN  = 2;
//
//  FUNCTION ASSIGNMENT:
//  R1 is CLEAN
//  R2 is LEAD
//  R3 is BOOST
//  R4 is REVERB
//
//  Map from footswitch LED State to MIDI Preset
//
//                                R
//                    C       B   E
//                    L   L   O   V
//                    E   E   O   E
//                    A   A   S   R
//                    N   D   T   B
//
//                   R1, R2, R3, R4
uint8_t REMOTE_PC_MAP [16] = {
  PRESET_RHYTHM,  //  0,  0,  0,  0
  PRESET_RHYTHM,  //  0,  0,  0,  1
  PRESET_RHYTHM,  //  0,  0,  1,  0
  PRESET_RHYTHM,  //  0,  0,  1,  1
  PRESET_LEAD,    //  0,  1,  0,  0
  PRESET_LEAD,    //  0,  1,  0,  1
  PRESET_LEAD,    //  0,  1,  1,  0
  PRESET_LEAD,    //  0,  1,  1,  1
  PRESET_CLEAN,   //  1,  0,  0,  0
  PRESET_CLEAN,   //  1,  0,  0,  1
  PRESET_CLEAN,   //  1,  0,  1,  0
  PRESET_CLEAN,   //  1,  0,  1,  1
  PRESET_LEAD,    //  1,  1,  0,  0
  PRESET_LEAD,    //  1,  1,  0,  1
  PRESET_LEAD,    //  1,  1,  1,  0
  PRESET_LEAD,    //  1,  1,  1,  1
};

#define MIDI_CHANNEL      0

#define REMOTE_TIMEOUT_MS 200   // Button debounce timeout

// LED pinout
#define LED_RED           3
#define LED_GREEN         2

bool force_update;              // Is true during the first loop execution right after RESET

void setup() {

  // Configure I/Os
  for (int i = 0; i < 4; i++) {
    pinMode(REMOTE_IN[i], INPUT);
    pinMode(REMOTE_OUT[i], OUTPUT);
    digitalWrite(REMOTE_OUT[i], LOW);
  }
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);

  // Flash LEDs
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, LOW);
  Serial.begin(31250);
  delay(1000);
  digitalWrite(LED_RED, LOW);
  digitalWrite(LED_GREEN, HIGH);

  // React to the current footswitch LED state right after RESET
  force_update = true;

}

void loop() {

  unsigned long now_ms = millis();

  bool any_timeout = false;
  
  for (int i = 0; i < 4; i++) {
    
    if (now_ms - REMOTE_LAST_MS[i] < REMOTE_TIMEOUT_MS) {

      // This switch is currently ignored (debounce period)
      any_timeout = true;

    } else {
      
      bool state = digitalRead(REMOTE_IN[i]);

      // Detect the R<x>_IN input state change
      if (REMOTE_STATE[i] != state || force_update) {

        force_update = false;

        // Save the current R<x>_IN input state and the current timestamp (needed to schedule debounce period)
        REMOTE_STATE[i] = state;
        REMOTE_LAST_MS[i] = now_ms;

        // Pass the R<x>_IN input state through to Laney Amp (Footswitch OUT)
        digitalWrite(REMOTE_OUT[i], state);

        // Construct the R<x>_IN input state vector
        uint8_t vector_input =
          (REMOTE_STATE[3] << 3) |  // REVERB
          (REMOTE_STATE[2] << 2) |  // BOOST
          (REMOTE_STATE[1] << 1) |  // LEAD
          (REMOTE_STATE[0] << 0);   // CLEAN
        
        // The footswitch LED state vector is the inverse of the R<x>_IN input state vector.
        uint8_t vector_led = ~vector_input & 0b00001111;

        // Address the Map using the footswitch LED state vector
        uint8_t midi_preset = REMOTE_PC_MAP[vector_led];

        // Send MIDI Program Change
        Serial.write(0xC0 | MIDI_CHANNEL);
        Serial.write(midi_preset);

        any_timeout = true;

      }
      
    }
  
  }

  // LED glows red during switch debounce timeout
  digitalWrite(LED_RED, any_timeout);

}
